% Kalman filter procedure to calcuate the PREDICTION, UPDATING and
% SMOOTHING equations and hence the mean bt|t and variance Pt|t of the state vector 
% and also draw the latent factor bt

% state - space is of the form
%    yt = Ht*bt + At*xt + et            et ~ N(0,R)
%    bt = mt + Ft*b(t-1)+ Gt*xt + ut    ut ~ N(0,Q)      
%    E(et,us) = 0  (but not necessarily, we could incorporate
%    correlation if we wanted to...)

% Prediction equations
%     bt|t-1 = mt + Ft*bt-1|t-1 + Gt*xt
%     Pt|t-1 = Ft*Pt-1|t-1*Ft' + Q
%     nt|t-1 = yt - yt|t-1 = yt - Ht*bt|t-1 - At*xt     PREDICTION ERROR
%     ft|t-1 = Ht*Pt|t-1*Ht' + R           CONDITIONAL VARIANCE
%     OF THE PREDICTION ERROR

% Updating equations
%     bt|t = bt|t-1 + Kt*nt|t-1
%     Pt|t = Pt|t-1 - Kt*Ht*Pt|t-1

% Kalman Gain
%     Kt = Pt|t-1*Ht'*(ft|t-1)^(-1)

% Smoothing equations (t = T-1, T-2, ... , 1)
%     bt|T = bt|t + Pt|t*Ft'*(P(t+1)|t)^(-1)*(b(t+1)|T - Ft*bt|t-mt)
%     Pt|T = Pt|t + Pt|t*Ft'*(P(t+1)|t)^(-1)*(P(t+1)|T -
%     P(t+1)|t)*(P(t+1)|t)^(-1)'*Ft*Pt|t'

function [bt, Pt, bt_T, Pt_T, ft_tlag]  = kalman_filter_aff(Yt, Ht, At, Xt, Rt, mt, Ft, Gt, Q, b0, P0); 

% INPUTS

%      Yt: is the observed data matrix of size  T x n 
%      where    n: number of observed variables, T: number of observations
%      Xt: is the observed exogenous variables matrix of size T x m  
%      where    m: number of observed exogenous variables

%      b0: Initial value for state vector bt of size nstates x 1 

%      Yt:     T x n 
%      Ht:     T x n x nstates
%      At:     n x m 
%      Xt:     T x m
%      Rt:     T x 1, R must be n x n  
%      mt:     nstates x 1
%      Ft:     nstates x  nstates
%      Gt:     nstates x m 
%      Q:      nstates x  nstates
%      b0:     nstates x 1  
%      P0:     nstates x nstates 

% OUTPUTS

%      bt:     unsmoothed Kalman beta estimates 
%      Pt:     unsmoothed Kalman covariance matrix estimates 
%      bt_T:   smoothed Kalman beta estimate 
%      Pt_T:   smoothed Kalman covariance matrix estimates 

T = size(Yt,1);
n = size(Yt,2);
m = size(Xt,2);
nstates = size(b0,1);

% denote bt|t-1 as bt_tlag   T    x  nstates
% denote bt|t   as bt_t      T+1  x  nstates
% denote Pt|t-1 as Pt_tlag   T    x  nstates  x  nstates 
% denote Pt|t   as Pt_t      T+1  x  nstates  x  nstates
% denote nt|t-1 as nt_tlag   T    x  n
% denote ft|t-1 as ft_tlag   T    x  n   x  n
% Kalman gain   Kt           T    x  nstates  x  n

%Initialize state vector bt
bt_tlag = zeros(T,nstates);
bt_t = zeros(T+1, nstates);
bt_t(1,:) = b0';

%Initialize state vector variance Pt
Pt_tlag = zeros(T, nstates, nstates);
Pt_t = zeros(T+1, nstates, nstates);
Pt_t(1,:,:) = P0;
%Initialize prediction error nt
nt_tlag = zeros(T, n);
%Initialize variance of the prediction error ft 
ft_tlag = zeros(T, n, n);
%Initialize Kalman gain vector
Kt = zeros(T, nstates, n);

% *********************************Running the Kalman Filter
for i=1:T
    R=Rt(i,:)*eye(n);
    % Prediction equations    
    if size(Xt,1)> 0
        bt_tlag(i,:) = (mt + Ft*bt_t(i,:)' + Gt*Xt(i,:)')';   
    else
        bt_tlag(i,:) = (mt + Ft*bt_t(i,:)')'; 
    end

    Pt_tlag(i,:,:) = Ft*reshape(Pt_t(i,:,:), nstates, nstates)*Ft' + Q;
    if size(Xt,1) > 0
        nt_tlag(i,:) = (Yt(i,:)' - reshape(Ht(i,:,:),n,nstates)*bt_tlag(i,:)' - At*Xt(i,:)')';   
    else
        nt_tlag(i,:) = (Yt(i,:)' - reshape(Ht(i,:,:),n,nstates)*bt_tlag(i,:)')';    
    end

    ft_tlag(i,:,:) = reshape(Ht(i,:,:),n,nstates)*reshape(Pt_tlag(i,:,:), nstates, nstates)*reshape(Ht(i,:,:),n,nstates)' + R;

    % Kalman gain
    Kt(i,:,:) = reshape(Pt_tlag(i,:,:), nstates, nstates)*reshape(Ht(i,:,:),n,nstates)'*inv(reshape(ft_tlag(i,:,:),n,n));

    % Updating equations
    bt_t(i+1, :) = (bt_tlag(i,:)' + reshape(Kt(i,:,:), nstates, n)*nt_tlag(i,:)')';  
    Pt_t(i+1,:,:) = ((reshape(Pt_tlag(i,:,:), nstates, nstates)- reshape(Kt(i,:,:),nstates,n)*reshape(Ht(i,:,:),n,nstates)*reshape(Pt_tlag(i,:,:),nstates,nstates)));
end    

bt = bt_t;
Pt = Pt_t;

% *******************************Running the Kalman Smoother
bt_t_ = bt_t(2:T+1,:);
Pt_t_ = Pt_t(2:T+1,:,:);

bt_T =  zeros(T, nstates);
bt_T(T,:) = bt_t_(T,:);
Pt_T =  zeros(T, nstates, nstates);
Pt_T(T,:,:) =  Pt_t_(T,:,:);

for i = 1:T-1
    tmp11 = bt_t_(T-i,:)';
    tmp12 = reshape(Pt_t_(T-i,:,:),nstates,nstates)*Ft'*inv(reshape(Pt_tlag(T-i+1,:,:),nstates,nstates));
    if size(Xt,1)>0
        tmp13 = (bt_T(T-i+1,:)'-Ft*bt_t_(T-i,:)' - mt - Gt*Xt(T-i,:)');
    else    
        tmp13 = (bt_T(T-i+1,:)'-Ft*bt_t_(T-i,:)' - mt);
    end
    bt_T(T-i,:) = (tmp11+tmp12*tmp13)';

    tmp1 = reshape(Pt_t_(T-i,:,:),nstates,nstates);
    tmp2 = reshape(Pt_t_(T-i,:,:),nstates,nstates)*Ft'*inv(reshape(Pt_tlag(T-i+1,:,:),nstates,nstates));
    tmp3 = reshape(Pt_T(T-i+1,:,:),nstates,nstates)-reshape(Pt_tlag(T-i+1,:,:),nstates,nstates);
    tmp4 =inv(reshape(Pt_tlag(T-i+1,:,:),nstates,nstates))'*Ft*reshape(Pt_t_(T-i,:,:),nstates,nstates)';
    Pt_T(T-i,:,:) = tmp1 + tmp2*tmp3*tmp4; 
end
